/*
*	refal.h - priskribo datumtipojn de la refalsistemo.
*	© 2005, A.A.Vladimirov.
*
*	Ĉi tiu verko estas libera programaro; vi povas ĝin pludistribui kaj/aŭ
*	modifi je la kondiĉoj de la GNUa Ĝenerala Publika Permesilo, eldonita de
*	«Free Software Foundation», laŭ la versio 2 de tiu Permesilo aŭ, se vi
*	preferas, ajna posta versio.
*
*	Ni distribuas ĉi tiun programon esperante ke ĝi estos utila, tamen
*	SEN IA AJN GARANTIO, i.a. sen la implica garantio pri SURMERKANTIGEBLO
*	aŭ TAŬGECO POR IU KONKRETA CELO. Pliajn detalojn vidu en la GNUa
*	Ĝenerala Publika Permesilo.
*
*	Ekzemplero de la GNUa Ĝenerala Publika Permesilo devas esti liverita al vi
*	kun ĉi tiu programo; se vi ĝin ne ricevis, turnu vin al
*	
*	Free Software Foundation, Inc.,
*	59 Temple Place, Suite 330,
*	Boston, MA 02111-1307
*	USA
*
*/

# include <stdio.h>
# include <stdlib.h>

/*
*	Rikordo LITERO estas datumtipo de litero, t.e. elemento de refalesprimo.
*	Tio estas la baza datumtipo de la refalsistemo.
*	
*	Speco de litero povas esti:
*
*	0) Cifero. Datumo estas entjero "cif" de tipo (unsigned int) kun valoro
*	inter 0 kaj 65535. 
*
*	1) Maldekstra literoida parentezo. Datumo estas referenco "dual" de tipo
*	(LITERO *) al responda dekstra parentezo.
*
*	2) Dekstra literoida parentezo. Datumo estas referenco "dual" de tipo
*	(LITERO *) al responda maldekstra parentezo.
*
*	# Literoidaj parentezoj emfazas sektoron de esprimo, kia estas simbolo.
*	# Sekve, kvanto de simbolojn en la lingvo estas potenciale malfinia.
*	# Ĉi tia maniero determini simbolojn estas konata en matematika logiko
*	# kiel "traduko": oni povas skribi literojn de arbitra alfabeto
*	# pere de vortoj el fiksita trilitera alfabeto "a()" - tiel nomataj
*	# 'literoidoj' (a), (aa), (aaa) k.t.p.
*
*	3) Maldekstra struktura parentezo. Datumo estas referenco "dual" de tipo
*	(LITERO *) al responda dekstra parentezo.
*
*	4) Dekstra struktura parentezo. Datumo estas referenco "dual" de tipo
*	(LITERO *) al responda maldekstra parentezo.
*
*	5) Konkretiganta litero "k". Datumo estas referenco "funk" de tipo
*	(FUNKCIO *) al aplikata funkcio.
*	
*	6) Konkretiganta punkto. Datumo estas referenco "dual" de tipo (LITERO *)
*	al responda konkretiganta litero "k".
*	
*	7) Variablo. Datumo estas referenco "var" de tipo (VARIABLO *) al tabelo
*	de parametroj de la variablo.
*
*/

typedef struct litero {
	int tip;				/* Datumtipo */
	union {

		unsigned int cif;		/* Cifero */
		struct litero *dual;	/* Referenco al duala parentezo */
		struct variablo *var;	/* Referenco al parametroj de variablo */
		struct funkcio *funk;	/* Referenco al funkcio */

	} dat;
	struct litero *ant;		/* Referenco al antaŭa litero */
	struct litero *sekv;	/* Referenco al sekva litero */
} LITERO;


# define CIFER 0			/* Cifero */
# define MD_LIT_PARENT 1		/* Maldekstra literoida parentezo */
# define D_LIT_PARENT 2		/* Dekstra literoida parentezo */
# define MD_STR_PARENT 3		/* Maldekstra struktura parentezo */
# define D_STR_PARENT 4		/* Dekstra struktura parentezo */
# define KONK_K 5			/* Konkretiganta litero "k" */
# define KONK_P 6			/* Konkretiganta punkto */
# define VAR 7				/* Variablo */

# define ESTAS_CIFER(x) (x->tip == 0)
# define ESTAS_MD_LIT_PARENT(x) (x->tip == 1)
# define ESTAS_D_LIT_PARENT(x) (x->tip == 2)
# define ESTAS_MD_STR_PARENT(x) (x->tip == 3)
# define ESTAS_D_STR_PARENT(x) (x->tip == 4)
# define ESTAS_KONK_K(x) (x->tip == 5)
# define ESTAS_KONK_P(x) (x->tip == 6)
# define ESTAS_VAR(x) (x->tip == 7)


/*
*	Rikordo VARIABLO estas datumtipo de tabelo de parametroj de variablo el
*	formuloj de substituo.
*	
*	Speco de variablo povas esti:
*	
*	0) Agema s-variablo. Tio estas la plej maldekstra variablo kun tia nomo
*	(en modelo), aŭ variablo, kies valoro estos kopio de iu sektoro - sed ne
*	estas sektoro mem! - el origina esprimo (en dekstra flanko). Valoro de
*	"ref" estas uzata nur por variablo el dekstra flanko: ĝi estas la
*	referenco al kopiinda variablo el modelo.
*
*	1) Malagema s-variablo. Tio estas variablo, maldekstre de kiu ekzistas
*	aliaj variabloj kun tia sama nomo (en modelo), aŭ variablo, kies valoro
*	estos iu sektoro - sed ne ĝia kopio! - el origina esprimo (en dekstra
*	flanko). Valoro de "ref" estas la referenco al la plej maldekstra el
*	variabloj kun tia sama nomo (por variablo el modelo), aŭ estas la
*	referenco al variablo, kies valoron oni devas transporti.
*	
*	2) Agema w-variablo. Tio estas la plej maldekstra variablo kun tia nomo
*	(en modelo), aŭ variablo, kies valoro estos kopio de iu sektoro - sed ne
*	estas sektoro mem! - el origina esprimo (en dekstra flanko). Valoro de
*	"ref" estas uzata nur por variablo el dekstra flanko: ĝi estas la
*	referenco al kopienda variablo el modelo.
*
*	3) Malagema w-variablo. Tio estas variablo, maldekstre de kiu ekzistas
*	aliaj variabloj kun tia sama nomo (en modelo), aŭ variablo, kies valoro
*	estos iu sektoro - sed ne ĝia kopio! - el origina esprimo (en dekstra
*	flanko). Valoro de "ref" estas la referenco al la plej maldekstra el
*	variabloj kun tia sama nomo (por variablo el modelo), aŭ estas la
*	referenco al variablo, kies valoron oni devas transporti.
*	
*	4) Agema e-variablo. Tio estas la plej maldekstra variablo kun tia nomo,
*	se ĝia valoro ne estas determinata per pozicio de ĝia maldekstra limo
*	(en modelo), aŭ variablo, kies valoro estos kopio de iu sektoro - sed ne
*	estas sektoro mem! - el origina esprimo (en dekstra flanko). Valoro de
*	"ref" estas uzata nur por variablo el dekstra flanko: ĝi estas la
*	referenco al kopienda variablo el modelo.
*	
*	5) Duonagema e-variablo. Tio estas la plej maldekstra variablo kun tia
*	nomo, se ĝia valoro estas determinata per pozicio de ĝia maldekstra limo
*	(en modelo). Valoro de "ref" estas ignorata.
*	
*	6) Malagema e-variablo. Tio estas variablo, maldekstre de kiu ekzistas
*	aliaj variabloj kun tia sama nomo (en modelo), aŭ variablo, kies valoro
*	estos iu sektoro - sed ne ĝia kopio! - el origina esprimo (en dekstra
*	flanko). Valoro de "ref" estas la referenco al la plej maldekstra el
*	variabloj kun tia sama nomo (por variablo el modelo), aŭ estas la
*	referenco al variablo, kies valoron oni devas transporti.
*	
*	Valoro de "ant" estas referenco al litero el refalseprimo, kiu antaŭas
*	valoron de la variablo. Valoro de "sekv" estas referenco al litero el
*	refalesprimo, kiu sekvas valoron de la variablo. Valoron de "ant" kaj
*	"sekv" estas trovenda dum alfrontado.
*	
*	# EKZEMPLO: Rigardu formulon de substituo
*	#
*	#		s/1/ (s/1/) e/2/ e/3/ = s/1/ e/2/ e/2/
*	#		1     2     3    4      5    6    7
*	#
*	# Variabloj 1 kaj 3 estas agemaj variabloj de modelo; variablo 4 estas
*	# duonagema; variablo 2 estas malagema, kaj responda agema variablo
*	# estas 1. Variabloj 5 kaj 6 estas malagemaj variabloj de dekstra
*	# flanko; por konstrui valoron de 5, oni povas preni valoron de 1
*	# (aŭ 2), kaj por konstrui valoron de 6, oni povas preni valoron de 3.
*	# Tamen variablo 7 estas agema ĉar variablo e2 aperas en la modelo nur
*	# unuope; por konstrui valoron de 7, oni igatas kopii valoron de 3.
*
*/

typedef struct variablo {
	int tip;				/* Speco de variablo */
	struct variablo *ref;	/* Referenco al alia variablo */
	struct litero *ant;		/* Referenco al la lasta litero antaŭ valoro */
	struct litero *sekv;	/* Referenco al la unua litero post valoro */
} VARIABLO;


# define S_AGEM 0
# define S_MALAGEM 1
# define W_AGEM 2
# define W_MALAGEM 3
# define E_AGEM 4
# define E_DUONAGEM 5
# define E_MALAGEM 6

# define ESTAS_S_AGEM(x) (x->tip == 0)
# define ESTAS_S_MALAGEM(x) (x->tip == 1)
# define ESTAS_W_AGEM(x) (x->tip == 2)
# define ESTAS_W_MALAGEM(x) (x->tip == 3)
# define ESTAS_E_AGEM(x) (x->tip == 4)
# define ESTAS_E_DUONAGEM(x) (x->tip == 5)
# define ESTAS_E_MALAGEM(x) (x->tip == 6)


/*
*	Rikordo SUBSTITUO estas datumtipo de formulo de substituo.
*
*	Valoro de "mod" estas referenco al la unua litero de modelo.
*	
*	# La unua litero de modelo devas esti konkretiganta litero "k",
*	# kaj la lasta litero de modelo devas esti konkretiganta punkto.
*	# Krome, speco de "dat" de la unua litero devas esti ne "funk", sed
*	# "dual" - la unua kaj la lasta literoj el modelo devas esti ligitaj.
*	
*	Valoro de "rez" estas referenco al la unua litero de dekstra flanko, aŭ
*	estas "NULL" (se dekstra flanko estas malplena).
*
*	# La lasta litero de nemalplena dekstra parto devas obei ekvacion
*	# "x->sekv == NULL".
*	
*	Valoro de "sekv" estas referenco al sekva formulo de substituo, aŭ
*	estas "NULL" (se la formulo estas la lasta).
*
*/


typedef struct substituo {
	struct litero *mod;		/* Referenco al modelo (maldekstra flanko) */
	struct litero *rez;		/* Referenco al dekstra flanko */
	struct substituo *sekv;	/* Referenco al sekva formulo de substituo */
} SUBSTITUO;


/*
*	Rikordo FUNKCIO estas datumtipo de funkcio de refalmaŝino.
*
*	Funkcio povas esti enkonstruita aŭ plene refala. Se funkcio estas
*	enkonstruita, valoro de "enk" estas 1; alie ĝi estas 0.
*
*	Enkonstruita funkcio estas skribita per lingvo C en formato:
*
*		void NOMO_DE_FUNKCIO (void);
*
*	Valoro de "dat.f" estas referenco al tia ĉi C-funkcio.
*
*	Plene refala funkcioj estas determinata per refalprogramo. Valoro de
*	"dat.s" estas referenco al unua ĉeno de listo de formuloj de substituo,
*	kiuj determinas la funkcion, aŭ estas "NULL", se la funkcio ne estas jam
*	determinita.
*
*	Ĉia funkcio (escepte de sola interna funkcio, haltanta refalmaŝinon)
*	havas nomon. Nomo de funkcio estas unikoda vorto; ĝi ŝparas kiel
*	tabelo de (unsigned long)-nombroj.
*
*	Funkcioj de la refalmaŝino formas liston. Kaŭze de tio valoro "sekv"
*	ekzistas: ĝi estas referenco al funkcio, kia sekvas (en listo de funkcio)
*	kurantan funkcion.
*	
*/


typedef struct funkcio {
	int enk;					/* Speco de funkcio */
	unsigned long *nom;			/* Nomo de funkcio */
	union {
	
		void (*f) (void);		/* Referenco al enkonstruita funkcio */
		struct substituo *s;	/* Referenco al unua FS */

	} dat;
	struct funkcio *sekv;		/* Referenco al sekva funkcio */
} FUNKCIO;


extern int ARGC;			/* Unua argumento de C-funkcio "main" */
extern char **ARGV;			/* Dua argumento de C-funkcio "main" */

extern int NOM_LIM;			/* Maksimuma longo de nomo */
extern int LIN_LIM;			/* Maksimuma longo de linio */
extern int VAR_LIM;			/* Maksimuma kvanto de variabloj en FS */

extern LITERO *LLM;			/* Listo de libera memoro */
extern FUNKCIO *LF;			/* Listo de funkcioj */
extern LITERO *LKP;			/* Listo de konkretigantaj punktoj */
extern LITERO *KKK;			/* Kuranta konkretiganta "k" */
extern LITERO *KKP;			/* Kuranta konkretiganta "." */
extern SUBSTITUO *KFS;		/* Kuranta formulo de substituo */


# define ligi_literojn(x,y) {x->sekv = y; y->ant = x;}
# define ligi_parentezojn(x,y) {x->dat.dual = y; y->dat.dual = x;}


# define krei_literon(x) {\
	x = LLM;\
	if (LLM->sekv == NULL) {\
		LLM = (LITERO *) malloc (sizeof (LITERO));\
		if (LLM == NULL) ref_eraro (1);\
		LLM->sekv = NULL;\
	}\
	else {\
		LLM = LLM->sekv;\
	};\
}
		

# define detrui_literon(x) {ligi_literojn (x, LLM); LLM = x;}

